/***************************************************************************
 *   Copyright (C) 2005-2006 by Georg Hennig                               *
 *   Email: georg.hennig@web.de                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _KFILEITEMEXT_H_
#define _KFILEITEMEXT_H_

#include <qmutex.h>

#include <kfileitem.h>

class KMD5;
class KURL;
class QString;
class KomparatorWidget;
class KTempFile;

enum CUSTOM_EVENT_TYPES
{
  RESULT_DUPLICATE = 12345,
  RESULT_COMP_MISSING,
  RESULT_COMP_NEWER,
  RESULT_COMP_NEWER_EQUAL,
  RESULT_COMP_DIFFERENT,
  RESULT_FINISHED,
  DOWNLOAD_FILE,
  DELETE_FILE,
  PROGRESS_STATUS,
  PROGRESS_PROGRESS
};

class KFileItemExt : public KFileItem
{
  public:
    KFileItemExt( const KIO::UDSEntry &_entry, const KURL &_url,
                  bool _determineMimeTypeOnDemand = false, bool _urlIsDirectory = false );
    KFileItemExt( mode_t _mode, mode_t _permissions, const KURL &_url,
                  bool _determineMimeTypeOnDemand = false );
    KFileItemExt( const KURL &url, const QString &mimeType, mode_t mode );
    KFileItemExt( const KFileItem &item );
    KFileItemExt( const KFileItemExt &item );
    ~KFileItemExt();

    QCString MD5( KomparatorWidget *parent );    // returns MD5 sum of the file. will be calculated
                                                 // only the first time and then stored in "md5".
    QString getFile( KomparatorWidget *parent ); // returns path to (temporary if remote) local file
    void deleteFile( KomparatorWidget *parent ); // deletes the temporary file if available.
                                                 // must be called after getFile(...).

    void setTmpFile( KTempFile *_tmpFile );
    void setDownloadError( int _download_error );

    int dir; // 0 left, 1 right

    int hasdupes; // true only if file is first on duplicate chain
    int isdupe;
    int isdupe_comp;
    KFileItemExt *duplicates;
    KFileItemExt *dup_parent; // parent in chain. NULL if it's the first item in chain.
    KFileItemExt *brother;    // a file in files_comp has a brother in files_dupes.
    KFileItemExt *next;       // next in chain.

  private:
    KMD5 *md5;
    KTempFile *tmpFile;
    QMutex mutex;
    int download_error;
};

#endif
