/***************************************************************************
 *   Copyright (C) 2005-2006 by Georg Hennig                               *
 *   Email: georg.hennig@web.de                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qcheckbox.h>
#include <qlabel.h>
#include <qradiobutton.h>
#include <qtabwidget.h>

#include <kcombobox.h>
#include <kconfig.h>
#include <kdialogbase.h>
#include <kglobal.h>
#include <kiconloader.h>
#include <klineedit.h>
#include <klocale.h>
#include <knuminput.h>
#include <kurlrequester.h>

#include "kdatecombo.h"
#include "komparatorwidget.h"
#include "komparatorpreset.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

KomparatorPreset::KomparatorPreset( KomparatorWidget *_parent )
{
  parent = _parent;
  preset_title = i18n( "default" );

  url1 = "";
  url2 = "";
  regex = "*";
  size_selection = 0; // 0 = none, 1 = at least, 2 = at most, 3 = equal to.
  size = 0; // spin box content
  size_multiple = 1; // 0 = bytes, 1 = KB, 2 = MB
  modified_or_created = 0; // 0 = any, 1 = between, 2 = during the last
  date1.setDate( QDate::currentDate() );
  date2.setDate( QDate::currentDate() );
  date3_num = 0;
  modified_last = 0; // 0 = minutes, 1 = hours, 2 = days, 3 = months, 4 = years
  ignore_empty = TRUE;
  size_comparison = TRUE;
  md5_sum = TRUE;
  binary_comparison = FALSE;
  binary_comparison_size = 10; // in MB
  empty_folders = FALSE;
  include_subfolders = TRUE;
  use_locate = FALSE;
  case_sensitive = TRUE;
  soft_links = FALSE;
  enable_duplicates_search = TRUE;
  enable_missing_search = TRUE;
  enable_newer_search = TRUE;
  reference_folder_left = TRUE;
}

QString KomparatorPreset::title()
{
  return preset_title;
}

bool KomparatorPreset::readConfig( uint preset_number )
{
  KConfig *conf = KGlobal::config();

  if ( !conf ) return FALSE;
  // This allows to create a KomparatorPreset, loadConfig and if it fails, the preset before was the last one.
  if ( !conf->hasGroup( QString( "Preset%1" ).arg( preset_number ) ) ) return FALSE;

  conf->setGroup( QString( "Preset%1" ).arg( preset_number ) );

  // Read the entries
  preset_title = conf->readEntry( "preset_title" );
  url1 = KURL::fromPathOrURL( conf->readEntry( "url1" ) );
  url2 = KURL::fromPathOrURL( conf->readEntry( "url2" ) );
  regex = conf->readEntry( "regex" );
  size_selection = conf->readNumEntry( "size_selection" ); // 0 = none, 1 = at least, 2 = at most, 3 = equal to.
  size = conf->readUnsignedLongNumEntry( "size" ); // spin box content
  size_multiple = conf->readUnsignedNumEntry( "size_multiple" ); // 0 = bytes, 1 = KB, 2 = MB
  modified_or_created = conf->readUnsignedNumEntry( "modified_or_created" ); // 0 = any, 1 = between, 2 = during the last
  date1 = conf->readDateTimeEntry( "date1" );
  date2 = conf->readDateTimeEntry( "date2" );
  date3_num = conf->readUnsignedNumEntry( "date3_num" );
  modified_last = conf->readUnsignedNumEntry( "modified_last" ); // 0 = minutes, 1 = hours, 2 = days, 3 = months, 4 = years
  ignore_empty = conf->readBoolEntry( "ignore_empty" );
  size_comparison = conf->readBoolEntry( "size_comparison" );
  md5_sum = conf->readBoolEntry( "md5_sum" );
  binary_comparison = conf->readBoolEntry( "binary_comparison" );
  binary_comparison_size = conf->readUnsignedNumEntry( "binary_comparison_size" ); // in MB
  empty_folders = conf->readBoolEntry( "empty_folders" );
  include_subfolders = conf->readBoolEntry( "include_subfolders" );
  use_locate = conf->readBoolEntry( "use_locate" );
  case_sensitive = conf->readBoolEntry( "case_sensitive" );
  soft_links = conf->readBoolEntry( "soft_links" );
  enable_duplicates_search = conf->readBoolEntry( "enable_duplicates_search" );
  enable_missing_search = conf->readBoolEntry( "enable_missing_search" );
  enable_newer_search = conf->readBoolEntry( "enable_newer_search" );
  reference_folder_left = conf->readBoolEntry( "reference_folder_left" );

  // Sanity check for values
  if ( regex.isEmpty() ) regex = "*";
  if ( size_selection > 3 ) size_selection = 0;
  if ( size_multiple > 2 ) size_multiple = 0;
  if ( modified_or_created > 2 ) modified_or_created = 0;
  if ( modified_last > 4 ) modified_last = 0;

  return TRUE;
}

bool KomparatorPreset::saveConfig( uint preset_number )
{
  KConfig *conf = KGlobal::config();

  if ( !conf ) return FALSE;

  conf->setGroup( QString( "Preset%1" ).arg( preset_number ) );

  conf->writeEntry( "preset_title", preset_title );
  conf->writeEntry( "url1", url1.pathOrURL() );
  conf->writeEntry( "url2", url2.pathOrURL() );
  conf->writeEntry( "regex", regex );
  conf->writeEntry( "size_selection", size_selection );
  conf->writeEntry( "size", size );
  conf->writeEntry( "size_multiple", size_multiple );
  conf->writeEntry( "modified_or_created", modified_or_created );
  conf->writeEntry( "date1", date1 );
  conf->writeEntry( "date2", date2 );
  conf->writeEntry( "date3_num", date3_num );
  conf->writeEntry( "modified_last", modified_last );
  conf->writeEntry( "ignore_empty", ignore_empty );
  conf->writeEntry( "size_comparison", size_comparison );
  conf->writeEntry( "md5_sum", md5_sum );
  conf->writeEntry( "binary_comparison", binary_comparison );
  conf->writeEntry( "binary_comparison_size", binary_comparison_size );
  conf->writeEntry( "empty_folders", empty_folders );
  conf->writeEntry( "include_subfolders", include_subfolders );
  conf->writeEntry( "use_locate", use_locate );
  conf->writeEntry( "case_sensitive", case_sensitive );
  conf->writeEntry( "soft_links", soft_links );
  conf->writeEntry( "enable_duplicates_search", enable_duplicates_search );
  conf->writeEntry( "enable_missing_search", enable_missing_search );
  conf->writeEntry( "enable_newer_search", enable_newer_search );
  conf->writeEntry( "reference_folder_left", reference_folder_left );

  return TRUE;
}

bool KomparatorPreset::readFromInterface( bool reread )
{
  KDialogBase *ask_name_dialog = new KDialogBase( parent, "ask_name_dialog", TRUE, i18n( "Add preset" ), KDialogBase::Ok | KDialogBase::Cancel );

  QVBox *page = ask_name_dialog->makeVBoxMainWidget();

  QLabel *preset_label = new QLabel( i18n( "Name of your new preset:" ), (QWidget*)page, "preset_label" );
  QLineEdit *preset_title_LineEdit = new QLineEdit( "", (QWidget*)page, "preset_title_LineEdit" );
  if ( reread )
  {
    preset_label->setText( i18n( "Name of your preset:" ) );
    preset_title_LineEdit->setReadOnly( TRUE );
    preset_title_LineEdit->setText( title() );
  }
  preset_title_LineEdit->setMinimumWidth( ask_name_dialog->fontMetrics().maxWidth()*20 );

  QLabel *url1_label = new QLabel( i18n( "Left URL:" ), (QWidget*)page, "url1_label" );
  QLineEdit *url1_LineEdit = new QLineEdit( parent->url1_KURLRequester->url(), (QWidget*)page, "url1_LineEdit" );
  url1_LineEdit->setMinimumWidth( ask_name_dialog->fontMetrics().maxWidth()*20 );
  if ( reread )
  {
    url1_label->setText( url1_label->text() + " (" + i18n( "Old URL:" ) + " \"" + url1.url() + "\" )" );
  }

  QLabel *url2_label = new QLabel( i18n( "Right URL:" ), (QWidget*)page, "url2_label" );
  QLineEdit *url2_LineEdit = new QLineEdit( parent->url2_KURLRequester->url(), (QWidget*)page, "url2_LineEdit" );
  url2_LineEdit->setMinimumWidth( ask_name_dialog->fontMetrics().maxWidth()*20 );
  if ( reread )
  {
    url2_label->setText( url2_label->text() + " (" + i18n( "Old URL:" ) + " \"" + url2.url() + "\" )" );
  }

  if ( ask_name_dialog->exec() == QDialog::Rejected )
  {
    delete ask_name_dialog;
    ask_name_dialog = NULL;
    return FALSE;
  }

  preset_title = preset_title_LineEdit->text();
  url1 = KURL::fromPathOrURL( url1_LineEdit->text() );
  url2 = KURL::fromPathOrURL( url2_LineEdit->text() );

  delete ask_name_dialog;
  ask_name_dialog = NULL;

  // Set entries from interface
  regex = parent->regex_KLineEdit->text();
  size_selection = parent->size_KComboBox->currentItem(); // 0 = none, 1 = at least, 2 = at most, 3 = equal to.
  size = parent->bytes_KIntNumInput->value(); // spin box content
  size_multiple = parent->bytes_KComboBox->currentItem(); // 0 = bytes, 1 = KB, 2 = MB
  modified_or_created = parent->modifiedNoRestriction_RadioButton->isChecked() ? 0 :
                        ( parent->modifiedBetween_RadioButton->isChecked() ? 1 : 2 ); // 0 = any, 1 = between, 2 = during the last
  QDate *tmp_date = new QDate;
  date1 = parent->modifiedFrom_KDateCombo->getDate( tmp_date );
  date2 = parent->modifiedTo_KDateCombo->getDate( tmp_date );
  date3_num = parent->modifiedLastTime_KIntNumInput->value();
  modified_last = parent->modifiedLastTime_KComboBox->currentItem(); // 0 = minutes, 1 = hours, 2 = days, 3 = months, 4 = years
  ignore_empty = parent->ignoreEmptyFiles_CheckBox->isChecked();
  size_comparison = parent->size_CheckBox->isChecked();
  md5_sum = parent->calculateChecksum_CheckBox->isChecked();
  binary_comparison = parent->binaryComparison_CheckBox->isChecked();
  binary_comparison_size = parent->binaryComparison_KIntSpinBox->value(); // in MB
  empty_folders = parent->findEmptyFolders_CheckBox->isChecked();
  include_subfolders = parent->includeSubfolders_CheckBox->isChecked();
  use_locate = parent->useFilesIndex_CheckBox->isChecked();
  case_sensitive = parent->caseSensitiveSearch_CheckBox->isChecked();
  soft_links = parent->followSoftLinks_CheckBox->isChecked();
  enable_duplicates_search = parent->isEnabled( DUPLICATES );
  enable_missing_search = parent->isEnabled( MISSING );
  enable_newer_search = parent->isEnabled( NEWER );
  reference_folder_left = parent->duplicatesReference1_RadioButton->isChecked();

  return TRUE;
}

void KomparatorPreset::setToInterface()
{
  // Set the entries in the GUI
  if ( !url1.isEmpty() ) parent->url1_KURLRequester->setKURL( url1 );
  if ( !url2.isEmpty() ) parent->url2_KURLRequester->setKURL( url2 );
  parent->regex_KLineEdit->setText( regex );
  parent->size_KComboBox->setCurrentItem( size_selection );
  parent->bytes_KIntNumInput->setValue( size );
  parent->bytes_KComboBox->setCurrentItem( size_multiple );
  parent->modifiedNoRestriction_RadioButton->setChecked( ( modified_or_created == 0 ) ? TRUE : FALSE );
  parent->modifiedBetween_RadioButton->setChecked( ( modified_or_created == 1 ) ? TRUE : FALSE );
  parent->modifiedLastTime_RadioButton->setChecked( ( modified_or_created == 2 ) ? TRUE : FALSE );
  parent->modifiedFrom_KDateCombo->setDate( date1.date() );
  parent->modifiedTo_KDateCombo->setDate( date2.date() );
  parent->modifiedLastTime_KIntNumInput->setValue( date3_num );
  parent->modifiedLastTime_KComboBox->setCurrentItem( modified_last );
  parent->ignoreEmptyFiles_CheckBox->setChecked( ignore_empty );
  parent->size_CheckBox->setChecked( size_comparison );
  parent->calculateChecksum_CheckBox->setChecked( md5_sum );
  parent->binaryComparison_CheckBox->setChecked( binary_comparison );
  parent->binaryComparison_KIntSpinBox->setValue( binary_comparison_size );
  parent->findEmptyFolders_CheckBox->setChecked( empty_folders );
  parent->includeSubfolders_CheckBox->setChecked( include_subfolders );
  parent->useFilesIndex_CheckBox->setChecked( use_locate );
  parent->caseSensitiveSearch_CheckBox->setChecked( case_sensitive );
  parent->followSoftLinks_CheckBox->setChecked( soft_links );
  parent->main_TabWidget->setTabIconSet( parent->main_TabWidget->page( 0 ),
      KGlobal::iconLoader()->loadIconSet( enable_duplicates_search ? "ok" : "no", KIcon::Toolbar, KIcon::SizeSmall ) );
  parent->main_TabWidget->setTabIconSet( parent->main_TabWidget->page( 1 ),
      KGlobal::iconLoader()->loadIconSet( enable_missing_search ? "ok" : "no", KIcon::Toolbar, KIcon::SizeSmall ) );
  parent->main_TabWidget->setTabIconSet( parent->main_TabWidget->page( 2 ),
      KGlobal::iconLoader()->loadIconSet( enable_newer_search ? "ok" : "no", KIcon::Toolbar, KIcon::SizeSmall ) );
  parent->duplicatesReference1_RadioButton->setChecked( reference_folder_left );
  parent->duplicatesReference2_RadioButton->setChecked( !reference_folder_left );
}
