/***************************************************************************
 *   Copyright (C) 2005-2006 by Georg Hennig                               *
 *   Email: georg.hennig@web.de                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef _KOMPARATORWIDGET_H_
#define _KOMPARATORWIDGET_H_

#include <kio/jobclasses.h>

#include "komparatorpreset.h"
#include "komparatorwidgetbase.h"

class KompareJob;
class KFileItem;
class KFileItemExt;
class QColor;

enum STATUS
{
  CLEAN = 0,
  SEARCHING_URL_1,
  SEARCHING_URL_2,
  COMPARING,
  READY
};

enum URL_MODE
{
  URL_URL = 0,
  URL_DIR_OR_URL,
  URL_DIR_OR_URL_ONLY,
  URL_FILE_NAME,
  URL_REPLACE_1,
  URL_REPLACE_2
};

enum LIST_VIEW
{
  DUPLICATES = 0,
  MISSING,
  NEWER
};

class KQuery;
class KPopupMenu;

class KomparatorWidget : public KomparatorWidgetBase
{
    Q_OBJECT

public:
    KomparatorWidget( const KURL &url1, const KURL &url2, QWidget* parent = 0, const char* name = 0, WFlags fl = 0 );
    ~KomparatorWidget();

    void keyPressEvent( QKeyEvent *event );

    void customEvent( QCustomEvent *event );

    KListView *currentListView; // public for key event handler.

    KompareJob *kompare;        // public for byte comparison (KFileItemExt::getFile(...))

    bool isEnabled( LIST_VIEW list_view ); // public for komparatorpreset

public slots:
    // USER INTERFACE SLOTS
    virtual void slotSearch();
    virtual void slotAbout();
    virtual void slotWhatsThis();
    virtual void slotKListViewDropped( KListView *list_view, QDropEvent *, QListViewItem * );
    virtual void slotContextMenu( KListView *list_view, QListViewItem *item, const QPoint &p );
    virtual void slotRemoveFromList();
    virtual void slotOpenBinding();
    virtual void slotExecute( QListViewItem *item );
    virtual void slotOpenFolder();
    virtual void slotCopySelection();
    virtual void slotTrashFiles();
    virtual void slotDeleteFiles();
    virtual void slotOpenWith();
    virtual void slotFileProperties();
    virtual void slotCopyNewerR2L();
    virtual void slotCopyNewerL2R();
    virtual void slotCopyMissingR2L();
    virtual void slotCopyMissingL2R();
    virtual void slotCopyToOtherSide();
    virtual void slotMoveMissingL2R();
    virtual void slotMoveMissingR2L();
    virtual void slotMoveNewerL2R();
    virtual void slotMoveNewerR2L();
    virtual void slotMoveToOtherSide();
    virtual void slotTrashDuplicates();
    virtual void slotDeleteDuplicates();
    virtual void slotRemoveSelectionMissing1();
    virtual void slotRemoveSelectionMissing2();
    virtual void slotRemoveSelectionNewer1();
    virtual void slotRemoveSelectionNewer2();
    virtual void slotReferenceLeft( bool is_ref );
    virtual void slotDuplicateSelected();
    virtual void slotPresetSelected( int sel );
    virtual void slotAddPreset();
    virtual void slotRemovePreset();
    virtual void slotUpdatePreset();

    // KQUERY SLOTS
    virtual void slotResult( int errorCode, QString errorString );
    virtual void slotAddFile( const KFileItem* item, const QString & );

    // KOMPAREJOB SLOTS
    virtual void slotFileDownloaded( KIO::Job *job );

    // INTERNAL SLOTS
    virtual void slotSizeTypeChanged( int i );
    virtual void slotMD5ChecksumEnabled( bool e );
    virtual void slotSizeEnabled( bool e );
    virtual void slotEmptyFolder( KIO::Job *job, const KIO::UDSEntryList &list );
    virtual void slotEmptyFolderFinished( KIO::Job *job );
    virtual void slotStatus( QString str );
    virtual void slotProgress( QString str, int percent );
    virtual void slotSaveListViewColumnWidth();

signals:
    void signalStatus( QString );
    void signalProgress( QString, int );
    void signalSaveListViewColumnWidth();

private:
    KQuery *query;
    STATUS status;

    QPtrList< KURL > symlinked_directories;

    KPopupMenu *popupMenu;

    KFileItemExt *files_dupes;
    KFileItemExt *files_comp;
    KFileItemExt *dirs_dupes;

    QStringList unreadable_files;
    uint number_of_files;

    int duplicates1_width[7];
    int duplicates2_width[8];
    int missing1_width[7];
    int missing2_width[7];
    int newer1_width[7];
    int newer2_width[7];

    QPtrList< KomparatorPreset > preset_list;

    QColor my_red, my_green, my_blue, my_pink;

    QIconSet broadIconSet( const QPixmap pm1, const QPixmap pm2, const QPixmap pm3 );

    void removeDeletedFromLists( KFileItemExt *item );
    void checkEmptyDirs( KURL::List &urls );

    bool initializeQuery();
    bool isDateValid();
    void enable( bool e );
    void resizeListViewColumns();

    QString prettyURL( const KURL &url, URL_MODE mode );
    void assumeValidURL( KURL &url );

    KFileItemExt *requested_download;
    KIO::Job *requested_download_job;
};

#endif

